/*
   netrik -- The ANTRIK Internet Viewer
   Copyright (C) Olaf D. Buddenhagen AKA antrik, et al (see AUTHORS)
   Published under the GNU GPL; see LICENSE for details.
*/
/*
 * layout.c -- contains the framework for the layouting module.
 *
 * (C) 2001 antrik
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cfg.h"
#include "debug.h"
#include "items.h"
#include "layout.h"
#include "links.h"
#include "load.h"
#include "render.h"
#include "syntax.h"
#include "url.h"
#include "url-history.h"

/*
 * load and layout a HTML page (prepare for rendering);
 *
 * Creates a descriptor for the page, calls layouting passes (syntax parsing,
 * element parsing, structure parsing, pre-rendering), prints debug messages in
 * between, and returns the number of the loaded page in the page list.
 *
 * The URL of the page to be loaded is determined by combining the "url" with
 * fallback values from "base_url" (allowing relative links etc.); if
 * "base_url" is NULL, default values are used as fallback. If "url" is NULL,
 * "base_url" is taken as the final url. (For already prepared URLs from
 * history.)
 *
 * The optional "form" argument points to a form item in the structure tree,
 * describing a form to submit with the HTTP request.
 *
 * When called with the a "reference" page (meaning the new page is identical
 * to this one, and only a link to a local anchor was followed), the page isn't
 * reloaded, and the layouting passes aren't performed; only a new descriptor
 * is created, and the data structures are copied from the reference.
 */

int load_page(base_url, url, form, reference, page_width, syntax_err)
const struct Url	*base_url;    /* fallback values */
const char		*url;    /* main URL */
const struct Item	*form;    /* form item of form to upload */
const struct Page	*reference;    /* old page; if not NULL, take layout data from it instead of reloading */
int			page_width;
enum Syntax_error	*syntax_err;    /* return: parse_syntax() found HTML syntax errors in page */
{
   struct Page		*page;    /* descriptor of processed page */

   if(reference==NULL) {    /* new document (not already in memory) -> load */
      if(url!=NULL) {    /* new page (not from history) */
	 int	del_page;
	 int	last;

	 for(last=page_list.pos; last>=0 && page_list.page[last]->url->proto.type==PT_INTERNAL; --last);    /* find last non-internal page in history */
	 page_list.pos=last+1;    /* add new page after last */

	 /* free descriptors of all pages lost from history by loading some new page when not at end (after going back, or when overwriting internal pages) */
	 DMSG(("freeing memory used by page descriptors lost from history\n"));
	 for(del_page=page_list.pos; del_page<page_list.count; ++del_page)
	    free_page(page_list.page[del_page]);

	 add_page();
      } else {    /* from history */
	 int	del_page;
	 int	last;

	 for(last=page_list.count-1; last>=0 && page_list.page[last]->url->proto.type==PT_INTERNAL; --last);    /* find last non-internal page in history */

	 DMSG(("freeing memory used by page descriptors of internal page...\n"));
	 for(del_page=last+1; del_page<page_list.count; ++del_page)
	    free_page(page_list.page[del_page]);
	 page_list.count=last+1;
      }
      page=page_list.page[page_list.pos];

      page->input=init_load(base_url, url, form);    /* open file/HTTP connection */

      DMSG(("parsing syntax...\n"));
      page->syntax_tree=parse_syntax(page->input, syntax_err);
      if(url!=NULL)    /* new page -> need to store effective URL */
	 page->url=page->input->url;
      uninit_load(page->input);    /* close file, free memory; also detects errors in wget loading */

#ifdef DEBUG
      if(cfg.debug) {
	 debug_printf("syntax tree:\n");
	 dump_tree(page->syntax_tree, 1, 0);    /* dump (including content, elements not parsed) */
      }
#endif

      DMSG(("parsing elements...\n"));
      parse_elements(page->syntax_tree);

#ifndef XHTML_ONLY
      DMSG(("fixing SGML (missing end tags)...\n"));
      sgml_rework(page->syntax_tree);
#ifdef DEBUG
      if(cfg.debug) {
	 debug_printf("fixed syntax tree:\n");
	 dump_tree(page->syntax_tree, 1, 1);    /* dump (including content, elements parsed) */
      }
#endif
#endif

      DMSG(("parsing structure...\n"));
      page->item_tree=parse_struct(page->syntax_tree);
      DMSG(("freeing syntax tree...\n"));
      free_syntax(page->syntax_tree, 1);    /* don't need syntax tree any longer */
      DMSG(("creating link list...\n"));
      page->links=make_link_list(page->item_tree);
      DMSG(("creating anchor list...\n"));
      page->anchors=make_anchor_list(page->item_tree);

      DMSG(("pre-rendering...\n"));
      page->page_map=pre_render(page->item_tree, page_width);
#ifdef DEBUG
      if(cfg.debug) {
	 debug_printf("item tree:\n");
	 dump_items(page->item_tree);
      }
#endif
   } else {    /* page already in memory -> copy layouting data */
      if(reference->active_link>=0)    /* some link is active in old page */
	 highlight_link(reference, reference->active_link, 1);    /* unhighlight (not active in new one) */
      
      page=page_list.page[page_list.pos];    /* assume page from history */
      if(url!=NULL) {    /* new page (not from history) -> create */
	 struct Url	*new_url;

	 new_url=merge_urls(base_url, url);
	 if(new_url->proto.type==PT_INTERNAL)    /* couldn't merge URLs */
	    return page_list.pos;    /* -> don't load anything */

	 ++page_list.pos;    /* add new after current page */
	 add_page();
	 page=page_list.page[page_list.pos];
	 page->url=new_url;
      }

      page->item_tree=reference->item_tree;
      page->links=reference->links;
      page->anchors=reference->anchors;
      page->page_map=reference->page_map;
   }

   if(*page->url->frag && url!=NULL) {    /* fragment identifier specified (and not from history) -> go to anchor */
      int	anchor;

      /* find desired anchor */
      for(anchor=0; anchor < page->anchors->count; ++anchor) {    /* all anchors in list */
	 const struct Item	*item=page->anchors->anchor_item[anchor];    /* item of currently tested anchor */
	 const char		*id=item->type==ITEM_BLOCK_ANCHOR?item->data.block_anchor->id:item->data.inline_anchor->id;    /* name of current anchor */

	 if(strcasecmp(id, page->url->frag)==0)    /* found right one -> don't search further */
	    break;
      }
      if(anchor < page->anchors->count)    /* found the anchor -> store */
	 page->active_anchor=anchor;
      else {
	 fprintf(stderr, "\nAnchor \"%s\" not found in page.\n", page->url->frag);
	 *syntax_err=1;
      }
   }    /* fragment identifier */

   return page_list.pos;
}

/* 
 * unallocate a page's layouting data
 *
 * Frees item tree and page map. (The input resource descriptor and the syntax
 * tree should already be cleared when calling this.)
 */
void free_layout_data(page)
const struct Page	*page;
{
   DMSG(("   freeing page map...\n"));
   free_map(page->item_tree, page->page_map);
   DMSG(("   freeing item tree...\n"));
   free_items(page->item_tree);
   DMSG(("   freeing link list...\n"));
   free_links(page->links);
   DMSG(("   freeing anchor list...\n"));
   free_anchors(page->anchors);
}

/*
 * unallocate page data
 *
 * Frees the split URL, and the page struct itself. (All other data should be
 * already cleared when calling this.)
 */
void free_page(page)
struct Page	*page;
{
   if(page->url!=NULL)
      free_url(page->url);

   DMSG(("   freeing page struct...\n"));
   free(page);
}
