/*
   netrik -- The ANTRIK Internet Viewer
   Copyright (C) Olaf D. Buddenhagen AKA antrik, et al (see AUTHORS)
   Published under the GNU GPL; see LICENSE for details.
*/
/*
 * main.c -- you guess it ;-) : Thats's the main program.
 *
 * (C) 2001, 2002 antrik
 *     2001, 2002 Patrice Neff
 *
 * handle startup (command line etc.); then load (and layout) first file (given
 * as argument), and either dump it and quit immediatly (with --dump), or
 * display it interactively and load new files (requested by user) in a loop
 */
#include <curses.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "config.h"

#include "cfg.h"
#include "cmdline.h"
#include "debug.h"
#include "forms.h"
#include "layout.h"
#include "links.h"
#include "pager.h"
#include "readline.h"
#include "render.h"
#include "screen.h"
#include "url.h"
#include "url-history.h"

int main(int argc, char *argv[])
{
   int			cur_page;    /* number of current page in page list */
   int			page_width;
   enum Pager_ret	pager_ret;    /* return value of pager (indicates if ended by 'q', by ':', or by following link) */
   int			opt_index;    /* position of first non-option parameter in command line */
   enum Syntax_error	syntax_err;    /* parse_syntax() found syntax errors in the HTML page */

/* Patrice --> */

   /* command line */
   opt_index=config_cmdln(argc, argv);
   if(argc-opt_index!=1) {    /* wrong number of non-option arguments */
      fprintf(stderr,"usage: %s html-file\n   or: %1$s -          (read from stdin)\n", argv[0]);
      exit(3);
   }

/* <-- Patrice */

   if(cfg.term_width)
      page_width=init_curses();
   else {
      init_curses();
      page_width=80;
   }

   cur_page=load_page(NULL, argv[opt_index], NULL, NULL, page_width, &syntax_err);    /* load URL given in command line argument */

   if(!cfg.dump) {
      if(syntax_err
#ifdef DEBUG
         || cfg.debug
#endif
      ) {
	 syntax_err=SE_NO;
	 fprintf(stderr, "\npress <return> to start viewer\n");
	 getchar();
      }
      start_curses();

      do {    /* until quit */
	 int	pager_wait=0;    /* wait for keypress before returning to pager */

         pager_ret=display(page_list.page[cur_page]);    /* start pager with current page (returns when 'q' pressed, when command line mode entered with ':', or when link followed) */

	 switch(pager_ret) {
	    /* pager ended by ':' */
	    case RET_COMMAND: {
	       char	*command;    /* user input */

	       command=readline(":");
	       if(command!=NULL && *command) {    /* nonempty (any command entered) */
		  add_history(command);
	       
		  if(strncmp(command, "e ", 2)==0 || strncmp(command, "E ", 2)==0) {    /* load file */
		     const char	*url=command+2;    /* name begins after "e " */

		     char	*blank_pos;    /* first blank in url */

		     /* remove trailing blank(s) */
		     blank_pos=strchr(url, ' ');    /* find first blank */
		     if(blank_pos!=NULL)    /* anything found */
			*blank_pos=0;    /* end string at first blank */

		     DMSG(("\nfreeing memory used by old page...\n"));
		     free_layout_data(page_list.page[cur_page]);    /* get rid of old page */

		     if(command[0]=='e' && page_list.page[cur_page]->url->proto.type!=PT_INTERNAL)    /* load relative URL */
			cur_page=load_page(page_list.page[cur_page]->url, url, NULL, NULL, page_width, &syntax_err);    /* load new page (use current page URL as base) */
		     else    /* load absolute URL */
			cur_page=load_page(NULL, url, NULL, NULL, page_width, &syntax_err);    /* load new page using absoulute URL (no base) */
		  } else {    /* command not "e ..." */
		     printf("unknown command\n");
		     pager_wait=1;
		  }    /* command not "e " */

#ifdef DEBUG
		  if(cfg.debug)
		     pager_wait=1;
#endif
	       }    /* command nonempty */
	       free(command);
	       break;
	    }    /* RET_COMMAND */

	    /* pager ended by following link */
	    case RET_LINK: {
	       const struct Page	*old_page=page_list.page[cur_page];
	       struct Link		*link=get_link(old_page, old_page->active_link);    /* link data in item tree */

	       switch(link->form) {    /* dispatch on link/form type */
		  case FORM_NO: {    /* normal link (not form control) */
		     char	url[strlen(link->value)+1];
		     strcpy(url, link->value);    /* save link URL before killing old page */

		     if(url[0]!='#') {    /* not local anchor -> load new page */
			DMSG(("\nfreeing memory used by old page...\n"));
			free_layout_data(old_page);    /* get rid of old page */
			cur_page=load_page(old_page->url, url, NULL, NULL, page_width, &syntax_err);    /* load link (use current page URL as base) */
		     } else    /* local anchor -> keep page, just jump to anchor */
			cur_page=load_page(old_page->url, url, NULL, old_page, page_width, &syntax_err);    /* load link (use current page URL as base) */
#ifdef DEBUG
		     if(cfg.debug)
			pager_wait=1;
#endif
		     break;
		  }    /* FORM_NO */

		  case FORM_TEXT:
		  case FORM_PASS: {
		     char	*new_value;

		     new_value=readline("value:");
		     if(new_value!=NULL) {    /* not aborted -> store new value */
			free(link->value);
			link->value=new_value;
			update_form(old_page, old_page->active_link);
		     }
		     break;
		  }

		  case FORM_CHECKBOX:
	          case FORM_MULTIOPTION:
		     link->enabled=1-link->enabled;
		     update_form(old_page, old_page->active_link);
		     break;

		  case FORM_RADIO:
		  case FORM_OPTION: {
		     const struct Item	*form_item=get_form_item(old_page, old_page->active_link);

		     struct Form_handle	handle=form_start(form_item, 0);    /* handle to iterate through all form controls (unfiltered) */
		     struct Link	*sibling;    /* other controls in this form */

		     /* disable any other radio buttons/options with same name */
		     if(link->name!=NULL)    /* belongs to some group at all... */
			while((sibling=form_next(&handle)) != NULL) {    /* for all form controls */
			   if(sibling->name!=NULL && !strcmp(sibling->name, link->name)) {    /* in same group -> disable */
			      if(sibling->enabled) {
				 sibling->enabled=0;
				 set_form(handle.cur_item->data.string, sibling);    /* beware -- crude hack! */
			      }
			   }
			}

		     link->enabled=1;
		     update_form(old_page, old_page->active_link);

		     break;
		  }

/* Patrice, antrik --> */
	          case FORM_SUBMIT: {		     
		     const struct Item	*form_item=get_form_item(old_page, old_page->active_link);

		     if(form_item==NULL) {    /* orphaned button (no parent form) */
			set_color_raw(COLOR_RED|8);
			printf("\nError: Submit button outside form\n");
			reset_color_raw();
			pager_wait=1;
			break;
		     }

		     if(form_item->data.form->url!=NULL) {    /* "action" given */
			if(form_item->data.form->method!=METHOD_GET)
			   continue;    /* POST not implemented yet... */

			link->enabled=1;    /* make the used submit button successful */

			cur_page=load_page(old_page->url, form_item->data.form->url, form_item, NULL, page_width, &syntax_err);    /* load form "action" URL (use current page URL as base); submit form data with HTTP request */
			DMSG(("\nfreeing memory used by old page...\n"));
			free_layout_data(old_page);
#ifdef DEBUG
			if(cfg.debug)
			   pager_wait=1;
#endif
		     } else {    /* no "action" */
			set_color_raw(COLOR_RED|8);
			printf("\nError: Form gives no URL to submit to\n");
			reset_color_raw();
			pager_wait=1;
		     }
		     break;
		  }    /* FORM_SUBMIT */
/* <-- Patrice, antrik */
		     
	       }    /* switch link type */

	       break;
	    }    /* RET_LINK */

	    /* "show link URL" requested */
	    case RET_LINK_URL: {
	       const struct Page	*page=page_list.page[cur_page];
	       const struct Link	*link=get_link(page, page->active_link);    /* link data in item tree */

	       if(!link->form) {    /* normal link (not form control) */
		  printf("\nlink URL:\n");
		  set_color_raw(COLOR_WHITE|8);
		  printf("%s\n", link->value);
		  reset_color_raw();
		  pager_wait=1;
	       }
	       break;
	    }    /* RET_LINK_URL */

	    /* show absolute link target URL */
	    case RET_ABSOLUTE_URL: {
	       const struct Page	*page=page_list.page[cur_page];
	       const struct Link	*link=get_link(page, page->active_link);    /* link data in item tree */

	       if(!link->form) {    /* normal link (not form control) */
		  struct Url	*target_url;    /* effective URL of link target */
		     
		  target_url=merge_urls(page->url, link->value);    /* (temporarily) construct target URL merged from current URL and link URL */
		  if(target_url->proto.type!=PT_INTERNAL) {    /* could extract absolute URL */
		     printf("\nabsolute link target URL:\n");
		     set_color_raw(COLOR_WHITE|8);
		     printf("%s\n", target_url->full_url);
		     reset_color_raw();
		  } else {
		     set_color_raw(COLOR_RED|8);
		     printf("\ncan't get target URL\n");
		     reset_color_raw();
		  }
		  free_url(target_url);    /* don't keep */

		  pager_wait=1;
	       }    /* normal link */
	       break;
	    }    /* RET_ABSOLUTE_URL */

	    /* "show current page URL" requested */
	    case RET_URL: {
	       const struct Page	*page=page_list.page[cur_page];
	       const char		*url_str=page->url->full_url;

	       if(page->url->proto.type!=PT_INTERNAL) {    /* has URL */
		  printf("\ncurrent page URL:\n");
		  set_color_raw(COLOR_WHITE|8);
		  printf("%s\n", url_str);
		  reset_color_raw();
	       } else    /* internal */
		  printf("\npage is from stdin and has no URL\n");
	       pager_wait=1;
	       break;
	    }    /* RET_LINK_URL */

	    /* pager ended by going back or forward in URL history */
	    case RET_HISTORY: {
	       struct Page	*ref=page_list.page[cur_page];    /* old page descriptor, if new page is from same URL (can reuse layout data) */
	       int		first_page, last_page;    /* range of page list entries to check for local */
	       int		page;    /* currently checked page list entry */

	       if(page_list.pos!=cur_page) {    /* not reloading same page -> check if can reuse layout data */
		  DMSG(("\nchecking whether we stay in same document...\n"));
		  if(page_list.pos>cur_page) {
		     first_page=cur_page;
		     last_page=page_list.pos;
		  } else {
		     first_page=page_list.pos;
		     last_page=cur_page;
		  }
		  for(page=first_page+1; page<=last_page; ++page) {    /* test all pages between old an new one for being created by jumping to local anchors */
		     if(!page_list.page[page]->url->local) {    /* not local -> can't reuse layout data */
			ref=NULL;
			break;
		     }
		  }
	       } else    /* reloading same page -> never reuse */
		  ref=NULL;

	       if(ref==NULL) {    /* layout data won't be reused */
		  DMSG(("freeing memory used by old page...\n"));
		  free_layout_data(page_list.page[cur_page]);
	       }

	       DMSG(("loading new page:\n"));
	       cur_page=load_page(page_list.page[page_list.pos]->url, NULL, NULL, ref, page_width, &syntax_err);    /* reload page from history (using already split url) */

#ifdef DEBUG
	       if(cfg.debug)
		  pager_wait=1;
#endif
	       break;
	    }    /* RET_HISTORY */

	    case RET_QUIT:
	       break;
	    case RET_NO:    /* suppress warning */
	 }    /* switch pager_ret */

	 if(syntax_err) {
	    syntax_err=SE_NO;
	    pager_wait=1;
	 }

	 if(pager_wait) {
	    printf("\npress <return> to return to pager\n"); fflush(stdout);
	    getch();
	 }

      } while(pager_ret!=RET_QUIT);    /* until pager ended by 'q' */
   } else {    /* cfg.dump */
      DMSG(("layouted page:\n"));
      dump(page_list.page[cur_page]);
   }

#ifdef DEBUG
   DMSG(("\nfreeing memory used by page...\n"));
   free_layout_data(page_list.page[cur_page]);

   DMSG(("freeing memory used by page list...\n"));
   free_page_list();
#endif

   return(0);
}
